
/*
 *   Link this into any client or server using NXConnection
 *   and Foundation. 
 */

#import <objc/Object.h>
#import <remote/NXProxy.h>
#import <foundation/NSValue.h>
#import <foundation/NSString.h>
#import <foundation/NSException.h>

#ifdef WIN32
#import <foundation/NSDistantObject.h>
#endif


/*
 *  This is the set of methods an OLE proxy talks over the wire
 */
@protocol NSIDispatchProxyProtocol

- (BOOL)respondsToSelector:(SEL)sel;
- (NSMethodSignature*) methodSignatureForSelector:(SEL)sel;
- (oneway void)setProtocolForNSIDispatchProxy:(Protocol*)proto;
- (struct objc_method_description*)descriptionForMethod:(SEL)sel;
- (struct objc_method_description*)methodDescriptionForSelector:(SEL)sel;
- (Class)classForPortCoder;
- (id)replacementObjectForPortCoder:(NSCoder *)aCoder;
- awakeAfterUsingCoder: (NSCoder*)aCoder;
- (void)_decrementReferencesBy:(int)i;

- (BOOL)isNSIDispatchProxy;
- (NSValue*)performInvocation:(NSInvocation*)inv getReturn:(char*)type;

- (NSString*)description;
- (void)setOLEPropertyNamed:(NSString*)name to:(NSValue*)val;
- (NSValue*)getOLEPropertyNamed:(NSString*)name;
- (NSEnumerator*)objectEnumerator;

@end



/*
 *  OLE Automation objects are represented as an 
 *  NXNSOLEPRoxy over an OLD DO NXConnection.  We need
 *  a few hacks to make this work right.
 */
@interface NXNSProxy : NXProxy @end
@interface NXNSOLEProxy : NXNSProxy 
@end
@interface NXNSOLEProxy (Private) 
- (void)setProtocolForNSIDispatchProxy:(Protocol*)p;
@end

/*
 *  This causes the protocol to be copied across the wire
 *  so the remote end knows what to expect.
 */
@implementation NXNSOLEProxy
- setProtocolForProxy:(Protocol*)p {
  id ret = [super setProtocolForProxy:p];
  [self setProtocolForNSIDispatchProxy:p]; // goes remote
  return ret; 
}

/*
 *  This method is only called from methodSignature:fromZone:
 */
- (struct objc_method_description*)descriptionForMethod:(SEL)op
{
	struct objc_method_description* desc, *desc2;

	desc2 = [@protocol(NSIDispatchProxyProtocol) descriptionForInstanceMethod:op];

	if (desc2)
	{
		desc = malloc (sizeof (*desc2));
		desc->types = malloc (strlen (desc2->types) + 1);
		desc->name = desc2->name;
		strcpy (desc->types, desc2->types);
		desc2 = desc;
	}
	
	return desc2;		
}
@end


/* 
 *	Our own little sizeof for Objective C types.
 *	Because we only pass the simple NSValue types by value
 *	this function only needs to know about the simple
 *	types.
 */
static inline unsigned NSSizeOfObjCType(const char *objCType) {
    switch (*objCType)
      {
      case 'i': case 'I':
        return sizeof (int);
      case 'l': case 'L':
        return sizeof (long);
      case 's': case 'S':
        return sizeof (short);
      case 'c': case 'C':
        return sizeof (char);
      case 'q': case 'Q':
        return sizeof (long long);
      case 'f':
        return sizeof (float);
      case 'd':	
        return sizeof (double);
      case '@':
        return sizeof (id);
      case ':':
        return sizeof (SEL);
      case '*': case '%':
        return sizeof (char*);
      default:
        [NSException raise:NSInternalInconsistencyException
                    format:@"Internal Error in %s", __FILE__];
      }
}


/*
 *  To copy NSValue's over an NXConnection we use this little helper
 *  class to get the memory management right.
 */
@interface NXDOTemporaryValueHolder : Object
{
   NSValue *value;
}
- initWithValue:(NSValue *)v;

@end

@interface NSValue (NXDOTemporaryValueHolder)
- encodeRemotelyFor:connection freeAfterEncoding:(BOOL *)flag isBycopy:(BOOL)isBycopy;
@end

@interface NSValue (NXDOTemporaryValueHolderPrivate)
-(void*)_value;
@end

@interface NSObject (NXDOTemporaryValueHolder)
- encodeRemotelyFor:connection freeAfterEncoding:(BOOL *)flag isBycopy:(BOOL)isBycopy;
@end


@implementation NXDOTemporaryValueHolder

- initWithValue:(NSValue *)v
{
     value = [v copy];
     return self;
}

- free
{
   [value release];
   return [super free];
}

- encodeUsing:portal
{
    const char *type = [value objCType];

    [portal encodeData:&type ofType:"*"];

    /* uses a private method ... ignore the warning */
    [portal encodeData:[value _value] ofType:type];

    return self;
}


- decodeUsing:portal
{
    char *type;
    void *valuePtr = nil;
    NSValue *newValue;

    [portal decodeData:&type ofType:"*"];

    /* maybe this should be alloca */
    valuePtr = malloc (NSSizeOfObjCType (type));
    [portal decodeData:valuePtr ofType:type];

    newValue = [NSValue value:valuePtr withObjCType:type];
    return newValue;
}

@end


@implementation NSValue (NXDOTemporaryValueHolder)
- encodeRemotelyFor:connection freeAfterEncoding:(BOOL *)flag isBycopy:(BOOL)isBycopy {

    switch (*[self objCType]) {
      case 'i': case 'I':
      case 'l': case 'L':
      case 's': case 'S':
      case 'c': case 'C':
      case 'q': case 'Q':
      case 'f': case 'd':    
      case '@': case ':':
      case '*': case '%':
        *flag = YES;
        return [[NXDOTemporaryValueHolder alloc] initWithValue:self];

      default:	
	/* This is defined in foundation, so don't worry... */
        return [super encodeRemotelyFor:connection
                      freeAfterEncoding:flag
                               isBycopy:isBycopy];
    }
}

@end



#ifdef WIN32
@interface NSDistantObject (PrivateAPI)
- awakeAfterUsingCoder:(NSCoder*)coder;
@end

@interface NSDistantIDispatchProxy : NSDistantObject
@end
	
@implementation NSDistantIDispatchProxy 

- awakeAfterUsingCoder:(NSCoder*)coder
{
	[self setProtocolForProxy:@protocol(NSIDispatchProxyProtocol)];
	return [super awakeAfterUsingCoder:coder];
}

- (struct objc_method_description*)methodDescriptionForSelector:(SEL)op
{
	return 0;
}

-(BOOL)isNSIDispatchProxy
{
	return YES;
}
 
@end
#endif


